/*
 * Copyright (C) 2010 Josh A. Beam
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h>
#include <GL/glut.h>

static const int WINDOW_WIDTH = 640;
static const int WINDOW_HEIGHT = 480;

/* scene functions defined in scene.c */
void sceneInit(void);
void sceneCycle(void);
void display(void);

static int g_window;

static GLuint g_program;
static GLuint g_programCameraPositionLocation;
static GLuint g_programLightPositionLocation;
static GLuint g_programLightColorLocation;
static GLuint g_cylinderBufferId;
static unsigned int g_cylinderNumVertices;
static float g_cameraPosition[3];

/* light properties */
#define NUM_LIGHTS 3
static float g_lightPosition[NUM_LIGHTS * 3];
static float g_lightColor[NUM_LIGHTS * 3];
static float g_lightRotation;


/* Returns a string containing the text in
 * a vertex/fragment shader source file.
 */
static char * shaderLoadSource(const char *filePath)
{
	const size_t blockSize = 512;
	FILE *fp;
	char buf[blockSize];
	char *source = NULL;
	size_t tmp, sourceLength = 0;

	/* open file */
	fp = fopen(filePath, "r");
	if(!fp) {
		fprintf(stderr, "shaderLoadSource(): Unable to open %s for reading\n", filePath);
		return NULL;
	}

	/* read the entire file into a string */
	while((tmp = fread(buf, 1, blockSize, fp)) > 0) {
		char *newSource = (char*) malloc(sourceLength + tmp + 1);
		if(!newSource) {
			fprintf(stderr, "shaderLoadSource(): malloc failed\n");
			if(source) free(source);
			return NULL;
		}

		if(source) {
			memcpy(newSource, source, sourceLength);
			free(source);
		}

		memcpy(newSource + sourceLength, buf, tmp);

		source = newSource;
		sourceLength += tmp;
	}

	/* close the file and null terminate the string */
	fclose(fp);
	if(source) source[sourceLength] = '\0';

	return source;
}

/*
 * Returns a shader object containing a shader
 * compiled from the given GLSL shader file.
 */
static GLuint shaderCompileFromFile(GLenum type, const char *filePath) {
	char *source;
	GLuint shader;
	GLint length, result;

	/* get shader source */
	source = shaderLoadSource(filePath);
	if(!source) return 0;

	/* create shader object, set the source, and compile */
	shader = glCreateShader(type);
	length = strlen(source);
	glShaderSource(shader, 1, (const char **)&source, &length);
	glCompileShader(shader);
	free(source);

	/* make sure the compilation was successful */
	glGetShaderiv(shader, GL_COMPILE_STATUS, &result);
	if(result == GL_FALSE) {
		char *log;

		/* get the shader info log */
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
		log = (char*) malloc(length);
		glGetShaderInfoLog(shader, length, &result, log);

		/* print an error message and the info log */
		fprintf(stderr, "shaderCompileFromFile(): Unable to compile %s: %s\n", filePath, log);
		free(log);

		glDeleteShader(shader);
		return 0;
	}

	return shader;
}

/*
 * Compiles and attaches a shader of the
 * given type to the given program object.
 */
void shaderAttachFromFile(GLuint program, GLenum type, const char *filePath)
{
	/* compile the shader */
	GLuint shader = shaderCompileFromFile(type, filePath);

	if(shader != 0) {
		/* attach the shader to the program */
		glAttachShader(program, shader);

		/* delete the shader - it won't actually be
		 * destroyed until the program that it's attached
		 * to has been destroyed */
		glDeleteShader(shader);
	}
}

static void createCylinder(unsigned int divisions) {
	const int floatsPerVertex = 6;
	unsigned int i, size;
	float *v;

	g_cylinderNumVertices = (divisions + 1) * 2;
	size = floatsPerVertex * g_cylinderNumVertices;

	/* generate vertex data */
	v = (float*) malloc(sizeof(float) * size);
	for(i = 0; i <= divisions; ++i) {
		float r = ((M_PI * 2.0f) / (float)divisions) * (float)i;
		unsigned int index1 = i * 2 * floatsPerVertex;
		unsigned int index2 = index1 + floatsPerVertex;

		/* vertex positions */
		v[index1 + 0] = cosf(r);
		v[index1 + 1] = 1.0f;
		v[index1 + 2] = -sinf(r);
		v[index2 + 0] = cosf(r);
		v[index2 + 1] = -1.0f;
		v[index2 + 2] = -sinf(r);

		/* normals */
		v[index1 + 3] = cosf(r);
		v[index1 + 4] = 0.0f;
		v[index1 + 5] = -sinf(r);
		v[index2 + 3] = v[index1 + 3];
		v[index2 + 4] = v[index1 + 4];
		v[index2 + 5] = v[index1 + 5];
	}

	/* create vertex buffer */
	glGenBuffers(1, &g_cylinderBufferId);
	glBindBuffer(GL_ARRAY_BUFFER, g_cylinderBufferId);
	glBufferData(GL_ARRAY_BUFFER, sizeof(float) * size, v, GL_STATIC_DRAW);
	free(v);

	/* enable arrays */
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	/* set pointers */
	glVertexPointer(3, GL_FLOAT, sizeof(float) * floatsPerVertex, 0);
	glNormalPointer(GL_FLOAT, sizeof(float) * floatsPerVertex, (const GLvoid *)(sizeof(float) * 3));
}

void sceneInit(void) {
	GLint result;

	/* create program object and attach shaders */
	g_program = glCreateProgram();
	shaderAttachFromFile(g_program, GL_VERTEX_SHADER, "shader.vp");
	shaderAttachFromFile(g_program, GL_FRAGMENT_SHADER, "shader.fp");

	/* link the program and make sure that there were no errors */
	glLinkProgram(g_program);

	glGetProgramiv(g_program, GL_LINK_STATUS, &result);

	if(result == GL_FALSE) {
		GLint length;
		char *log;

		/* get the program info log */
		glGetProgramiv(g_program, GL_INFO_LOG_LENGTH, &length);
		log = (char*) malloc(length);
		glGetProgramInfoLog(g_program, length, &result, log);

		/* print an error message and the info log */
		fprintf(stderr, "sceneInit(): Program linking failed: %s\n", log);
		free(log);

		/* delete the program */
		glDeleteProgram(g_program);
		g_program = 0;
	}

	/* get uniform locations */
	g_programCameraPositionLocation = glGetUniformLocation(g_program, "cameraPosition");
	g_programLightPositionLocation = glGetUniformLocation(g_program, "lightPosition");
	g_programLightColorLocation = glGetUniformLocation(g_program, "lightColor");

	/* set up red/green/blue lights */
	g_lightColor[0] = 1.0f;
	g_lightColor[1] = 0.0f;
	g_lightColor[2] = 0.0f;
	g_lightColor[3] = 0.0f;
	g_lightColor[4] = 1.0f;
	g_lightColor[5] = 0.0f;
	g_lightColor[6] = 0.0f;
	g_lightColor[7] = 0.0f;
	g_lightColor[8] = 1.0f;

	/* create cylinder */
	createCylinder(36);

	/* do the first cycle to initialize positions */
	g_lightRotation = 0.0f;

	sceneCycle();

	/* setup camera */
	g_cameraPosition[0] = 0.0f;
	g_cameraPosition[1] = 0.0f;
	g_cameraPosition[2] = 4.0f;
	glLoadIdentity();
	glTranslatef(-g_cameraPosition[0], -g_cameraPosition[1], -g_cameraPosition[2]);

}

void display(void) {

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* enable program and set uniform variables */
	glUseProgram(g_program);
	glUniform3fv(g_programCameraPositionLocation, 1, g_cameraPosition);
	glUniform3fv(g_programLightPositionLocation, NUM_LIGHTS, g_lightPosition);
	glUniform3fv(g_programLightColorLocation, NUM_LIGHTS, g_lightColor);

	/* render the cylinder */
	glDrawArrays(GL_TRIANGLE_STRIP, 0, g_cylinderNumVertices);

	/* disable program */
	glUseProgram(0);

	/* render each light */
	for(int i = 0; i < NUM_LIGHTS; ++i) {
		/* render sphere with the light's color/position */
		glPushMatrix();
		glTranslatef(g_lightPosition[i * 3 + 0], g_lightPosition[i * 3 + 1], g_lightPosition[i * 3 + 2]);
		glColor3fv(g_lightColor + (i * 3));
		glutSolidSphere(0.04, 36, 36);
		glPopMatrix();
	}

	glutSwapBuffers();
}

static unsigned int getTicks(void)
{
#ifdef _WIN32
	return GetTickCount();
#else
	struct timeval t;

	gettimeofday(&t, NULL);

	return (t.tv_sec * 1000) + (t.tv_usec / 1000);
#endif /* _WIN32 */
}

void sceneCycle(void) {
	static unsigned int prevTicks = 0;
	unsigned int ticks;
	float secondsElapsed;
	int i;

	/* calculate the number of seconds that have
	 * passed since the last call to this function */
	if(prevTicks == 0) prevTicks = getTicks();

	ticks = getTicks();
	secondsElapsed = (float)(ticks - prevTicks) / 1000.0f;
	prevTicks = ticks;

	/* update the light positions */
	g_lightRotation += (M_PI / 4.0f) * secondsElapsed;

	for(i = 0; i < NUM_LIGHTS; ++i) {
		const float radius = 1.75f;
		float r = (((M_PI * 2.0f) / (float)NUM_LIGHTS) * (float)i) + g_lightRotation;

		g_lightPosition[i * 3 + 0] = cosf(r) * radius;
		g_lightPosition[i * 3 + 1] = cosf(r) * sinf(r);
		g_lightPosition[i * 3 + 2] = sinf(r) * radius;
	}

	glutPostRedisplay();
}

static void handleKeyPress(unsigned char key, int x, int y)
{
	switch(key) {
		default:
			break;
		case 27: /* escape */
			glutDestroyWindow(g_window);
			exit(0);
			break;
	}
}

static void setPerspective(float fov1, float aspect1, float near1, float far1)
{
	float f;
	float mat[16];

	f = 1.0f / tanf(fov1 / 2.0f);

	mat[0] = f / aspect1;
	mat[1] = 0.0f;
	mat[2] = 0.0f;
	mat[3] = 0.0f;

	mat[4] = 0.0f;
	mat[5] = f;
	mat[6] = 0.0f;
	mat[7] = 0.0f;

	mat[8] = 0.0f;
	mat[9] = 0.0f;
	mat[10] = (far1 + near1) / (near1  - far1);
	mat[11] = -1.0f;

	mat[12] = 0.0f;
	mat[13] = 0.0f;
	mat[14] = (2.0f * far1 * near1) / (near1 - far1);
	mat[15] = 0.0f;

	glMultMatrixf(mat);
}

void init() {
	/* setup initial GL settings */
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0f);
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);

	/* set projection matrix */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	setPerspective(M_PI / 4.0f, (float)WINDOW_WIDTH / (float)WINDOW_HEIGHT, 0.1f, 200.0f);
	glMatrixMode(GL_MODELVIEW);

	/* initialize scene and start GLUT main loop */
	sceneInit();

}

int main(int argc, char *argv[]) {
	/* initialize GLUT */
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
	g_window = glutCreateWindow("OpenGL demo - http://joshbeam.com/");
    glewInit();
	init();
	/* setup callbacks */
	glutDisplayFunc(display);
	glutIdleFunc(sceneCycle);
	glutKeyboardFunc(handleKeyPress);
	glutMainLoop();
	return 0;
}
